"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../../sucrase-babylon/tokenizer/types");
const Transformer_1 = require("./Transformer");
class JSXTransformer extends Transformer_1.default {
    constructor(rootTransformer, tokens, importProcessor, nameManager, filePath) {
        super();
        this.rootTransformer = rootTransformer;
        this.tokens = tokens;
        this.importProcessor = importProcessor;
        this.nameManager = nameManager;
        this.filePath = filePath;
        this.lastLineNumber = 1;
        this.lastIndex = 0;
        this.filenameVarName = null;
    }
    process() {
        if (this.tokens.matches1(types_1.types.jsxTagStart)) {
            this.processJSXTag();
            return true;
        }
        return false;
    }
    getPrefixCode() {
        if (this.filenameVarName) {
            return `const ${this.filenameVarName} = ${JSON.stringify(this.filePath || "")};`;
        }
        else {
            return "";
        }
    }
    /**
     * Lazily calculate line numbers to avoid unneeded work. We assume this is always called in
     * increasing order by index.
     */
    getLineNumberForIndex(index) {
        const code = this.tokens.code;
        while (this.lastIndex < index && this.lastIndex < code.length) {
            if (code[this.lastIndex] === "\n") {
                this.lastLineNumber++;
            }
            this.lastIndex++;
        }
        return this.lastLineNumber;
    }
    getFilenameVarName() {
        if (!this.filenameVarName) {
            this.filenameVarName = this.nameManager.claimFreeName("_jsxFileName");
        }
        return this.filenameVarName;
    }
    processProps(firstTokenStart) {
        const lineNumber = this.getLineNumberForIndex(firstTokenStart);
        const devProps = `__self: this, __source: {fileName: ${this.getFilenameVarName()}, lineNumber: ${lineNumber}}`;
        if (!this.tokens.matches1(types_1.types.jsxName) && !this.tokens.matches1(types_1.types.braceL)) {
            this.tokens.appendCode(`, {${devProps}}`);
            return;
        }
        this.tokens.appendCode(`, {`);
        while (true) {
            if (this.tokens.matches2(types_1.types.jsxName, types_1.types.eq)) {
                if (this.tokens.currentToken().value.includes("-")) {
                    this.tokens.replaceToken(`'${this.tokens.currentToken().value}'`);
                }
                else {
                    this.tokens.copyToken();
                }
                this.tokens.replaceToken(": ");
                if (this.tokens.matches1(types_1.types.braceL)) {
                    this.tokens.replaceToken("");
                    this.rootTransformer.processBalancedCode();
                    this.tokens.replaceToken("");
                }
                else {
                    this.processStringPropValue();
                }
            }
            else if (this.tokens.matches1(types_1.types.jsxName)) {
                this.tokens.copyToken();
                this.tokens.appendCode(": true");
            }
            else if (this.tokens.matches1(types_1.types.braceL)) {
                this.tokens.replaceToken("");
                this.rootTransformer.processBalancedCode();
                this.tokens.replaceToken("");
            }
            else {
                break;
            }
            this.tokens.appendCode(",");
        }
        this.tokens.appendCode(` ${devProps}}`);
    }
    processStringPropValue() {
        const value = this.tokens.currentToken().value;
        const replacementCode = formatJSXTextReplacement(value);
        const literalCode = formatJSXStringValueLiteral(value);
        this.tokens.replaceToken(literalCode + replacementCode);
    }
    /**
     * Process the first part of a tag, before any props.
     */
    processTagIntro() {
        // Walk forward until we see one of these patterns:
        // jsxName to start the first prop, preceded by another jsxName to end the tag name.
        // [open brace] to start the first prop.
        // [jsxTagEnd] to end the open-tag.
        // [slash, jsxTagEnd] to end the self-closing tag.
        let introEnd = this.tokens.currentIndex() + 1;
        while (!this.tokens.matchesAtIndex(introEnd - 1, ["jsxName", "jsxName"]) &&
            !this.tokens.matchesAtIndex(introEnd, ["{"]) &&
            !this.tokens.matchesAtIndex(introEnd, ["jsxTagEnd"]) &&
            !this.tokens.matchesAtIndex(introEnd, ["/", "jsxTagEnd"])) {
            introEnd++;
        }
        if (introEnd === this.tokens.currentIndex() + 1 &&
            startsWithLowerCase(this.tokens.currentToken().value)) {
            this.tokens.replaceToken(`'${this.tokens.currentToken().value}'`);
        }
        while (this.tokens.currentIndex() < introEnd) {
            this.rootTransformer.processToken();
        }
    }
    processChildren() {
        while (true) {
            if (this.tokens.matches2(types_1.types.jsxTagStart, types_1.types.slash)) {
                // Closing tag, so no more children.
                return;
            }
            if (this.tokens.matches1(types_1.types.braceL)) {
                if (this.tokens.matches2(types_1.types.braceL, types_1.types.braceR)) {
                    // Empty interpolations and comment-only interpolations are allowed
                    // and don't create an extra child arg.
                    this.tokens.replaceToken("");
                    this.tokens.replaceToken("");
                }
                else {
                    // Interpolated expression.
                    this.tokens.replaceToken(", ");
                    this.rootTransformer.processBalancedCode();
                    this.tokens.replaceToken("");
                }
            }
            else if (this.tokens.matches1(types_1.types.jsxTagStart)) {
                // Child JSX element
                this.tokens.appendCode(", ");
                this.processJSXTag();
            }
            else if (this.tokens.matches1(types_1.types.jsxText)) {
                this.processChildTextElement();
            }
            else {
                throw new Error("Unexpected token when processing JSX children.");
            }
        }
    }
    processChildTextElement() {
        const value = this.tokens.currentToken().value;
        const replacementCode = formatJSXTextReplacement(value);
        const literalCode = formatJSXTextLiteral(value);
        if (literalCode === '""') {
            this.tokens.replaceToken(replacementCode);
        }
        else {
            this.tokens.replaceToken(`, ${literalCode}${replacementCode}`);
        }
    }
    processJSXTag() {
        const resolvedReactName = this.importProcessor.getIdentifierReplacement("React") || "React";
        const firstTokenStart = this.tokens.currentToken().start;
        // First tag is always jsxTagStart.
        this.tokens.replaceToken(`${resolvedReactName}.createElement(`);
        this.processTagIntro();
        this.processProps(firstTokenStart);
        if (this.tokens.matches2(types_1.types.slash, types_1.types.jsxTagEnd)) {
            // Self-closing tag.
            this.tokens.replaceToken("");
            this.tokens.replaceToken(")");
        }
        else if (this.tokens.matches1(types_1.types.jsxTagEnd)) {
            this.tokens.replaceToken("");
            // Tag with children.
            this.processChildren();
            while (!this.tokens.matches1(types_1.types.jsxTagEnd)) {
                this.tokens.replaceToken("");
            }
            this.tokens.replaceToken(")");
        }
        else {
            throw new Error("Expected either /> or > at the end of the tag.");
        }
    }
}
exports.default = JSXTransformer;
function startsWithLowerCase(s) {
    return s[0] === s[0].toLowerCase();
}
/**
 * Turn the given jsxText string into a JS string literal. Leading and trailing
 * whitespace on lines is removed, except immediately after the open-tag and
 * before the close-tag. Empty lines are completely removed, and spaces are
 * added between lines after that.
 *
 * We use JSON.stringify to introduce escape characters as necessary, and trim
 * the start and end of each line and remove blank lines.
 */
function formatJSXTextLiteral(text) {
    let result = "";
    let whitespace = "";
    let isInInitialLineWhitespace = false;
    let seenNonWhitespace = false;
    for (const c of text) {
        if (c === " " || c === "\t") {
            if (!isInInitialLineWhitespace) {
                whitespace += c;
            }
        }
        else if (c === "\n") {
            whitespace = "";
            isInInitialLineWhitespace = true;
        }
        else {
            if (seenNonWhitespace && isInInitialLineWhitespace) {
                result += " ";
            }
            result += whitespace;
            whitespace = "";
            result += c;
            seenNonWhitespace = true;
            isInInitialLineWhitespace = false;
        }
    }
    if (!isInInitialLineWhitespace) {
        result += whitespace;
    }
    return JSON.stringify(result);
}
/**
 * Produce the code that should be printed after the JSX text string literal,
 * with most content removed, but all newlines preserved and all spacing at the
 * end preserved.
 */
function formatJSXTextReplacement(text) {
    let numNewlines = 0;
    let numSpaces = 0;
    for (const c of text) {
        if (c === "\n") {
            numNewlines++;
            numSpaces = 0;
        }
        else if (c === " ") {
            numSpaces++;
        }
    }
    return "\n".repeat(numNewlines) + " ".repeat(numSpaces);
}
/**
 * Format a string in the value position of a JSX prop.
 *
 * Use the same implementation as convertAttribute from
 * babel-helper-builder-react-jsx.
 */
function formatJSXStringValueLiteral(text) {
    return JSON.stringify(text.replace(/\n\s+/g, " "));
}
