"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../../sucrase-babylon/tokenizer/types");
const isIdentifier_1 = require("../util/isIdentifier");
const Transformer_1 = require("./Transformer");
class TypeScriptTransformer extends Transformer_1.default {
    constructor(rootTransformer, tokens) {
        super();
        this.rootTransformer = rootTransformer;
        this.tokens = tokens;
    }
    process() {
        const processedType = this.rootTransformer.processPossibleTypeRange();
        if (processedType) {
            return true;
        }
        if (this.tokens.matches1(types_1.types._public) ||
            this.tokens.matches1(types_1.types._protected) ||
            this.tokens.matches1(types_1.types._private) ||
            this.tokens.matches1(types_1.types._abstract) ||
            this.tokens.matches1(types_1.types._readonly) ||
            this.tokens.matches1(types_1.types.nonNullAssertion)) {
            this.tokens.removeInitialToken();
            return true;
        }
        if (this.tokens.matches1(types_1.types._enum) || this.tokens.matches2(types_1.types._const, types_1.types._enum)) {
            this.processEnum();
            return true;
        }
        if (this.tokens.matches2(types_1.types._export, types_1.types._enum) ||
            this.tokens.matches3(types_1.types._export, types_1.types._const, types_1.types._enum)) {
            this.processEnum(true);
            return true;
        }
        return false;
    }
    processEnum(isExport = false) {
        // We might have "export const enum", so just remove all relevant tokens.
        this.tokens.removeInitialToken();
        while (this.tokens.matches1(types_1.types._const) || this.tokens.matches1(types_1.types._enum)) {
            this.tokens.removeToken();
        }
        const enumName = this.tokens.currentToken().value;
        this.tokens.removeToken();
        this.tokens.appendCode(`var ${enumName}; (function (${enumName})`);
        this.tokens.copyExpectedToken("{");
        this.processEnumBody(enumName);
        this.tokens.copyExpectedToken("}");
        if (isExport) {
            this.tokens.appendCode(`)(${enumName} || (exports.${enumName} = ${enumName} = {}));`);
        }
        else {
            this.tokens.appendCode(`)(${enumName} || (${enumName} = {}));`);
        }
    }
    /**
     * Rather than try to compute the actual enum values at compile time, we just create variables for
     * each one and let everything evaluate at runtime. There's some additional complexity due to
     * handling string literal names, including ones that happen to be valid identifiers.
     */
    processEnumBody(enumName) {
        let isPreviousValidIdentifier = false;
        let lastValueReference = null;
        while (true) {
            if (this.tokens.matches1(types_1.types.braceR)) {
                break;
            }
            const nameToken = this.tokens.currentToken();
            let name;
            let isValidIdentifier;
            let nameStringCode;
            if (nameToken.type.label === "name") {
                name = nameToken.value;
                isValidIdentifier = true;
                nameStringCode = `"${name}"`;
            }
            else if (nameToken.type.label === "string") {
                name = nameToken.value;
                isValidIdentifier = isIdentifier_1.default(name);
                nameStringCode = this.tokens.code.slice(nameToken.start, nameToken.end);
            }
            else {
                throw new Error("Expected name or string at beginning of enum element.");
            }
            this.tokens.removeInitialToken();
            let valueIsString;
            let valueCode;
            if (this.tokens.matches1(types_1.types.eq)) {
                const rhsEndIndex = this.tokens.currentToken().rhsEndIndex;
                if (rhsEndIndex == null) {
                    throw new Error("Expected rhsEndIndex on enum assign.");
                }
                this.tokens.removeToken();
                if (this.tokens.matches2(types_1.types.string, types_1.types.comma) ||
                    this.tokens.matches2(types_1.types.string, types_1.types.braceR)) {
                    valueIsString = true;
                }
                const startToken = this.tokens.currentToken();
                while (this.tokens.currentIndex() < rhsEndIndex) {
                    this.tokens.removeToken();
                }
                valueCode = this.tokens.code.slice(startToken.start, this.tokens.tokenAtRelativeIndex(-1).end);
            }
            else {
                valueIsString = false;
                if (lastValueReference != null) {
                    if (isPreviousValidIdentifier) {
                        valueCode = `${lastValueReference} + 1`;
                    }
                    else {
                        valueCode = `(${lastValueReference}) + 1`;
                    }
                }
                else {
                    valueCode = "0";
                }
            }
            if (this.tokens.matches1(types_1.types.comma)) {
                this.tokens.removeToken();
            }
            let valueReference;
            if (isValidIdentifier) {
                this.tokens.appendCode(`const ${name} = ${valueCode}; `);
                valueReference = name;
            }
            else {
                valueReference = valueCode;
            }
            if (valueIsString) {
                this.tokens.appendCode(`${enumName}[${nameStringCode}] = ${valueReference};`);
            }
            else {
                this.tokens.appendCode(`${enumName}[${enumName}[${nameStringCode}] = ${valueReference}] = ${nameStringCode};`);
            }
            lastValueReference = valueReference;
            isPreviousValidIdentifier = isValidIdentifier;
        }
    }
}
exports.default = TypeScriptTransformer;
